use std::{ptr, str};
use std::fmt::{Formatter, LowerHex, UpperHex};
use crate::{Uuid};

pub const UUID_STR_LEN: usize = 36;
const LOWER_LUT: [u8; 16] = [
    b'0', b'1', b'2', b'3', b'4', b'5', b'6', b'7', b'8', b'9', b'a', b'b', b'c', b'd', b'e', b'f'
];
const UPPER_LUT: [u8; 16] = [
    b'0', b'1', b'2', b'3', b'4', b'5', b'6', b'7', b'8', b'9', b'A', b'B', b'C', b'D', b'E', b'F'
];
const GROUPS: [(usize, usize); 5] = [
    (0, 8), (9, 13), (14, 18), (19, 23), (24, 36)
];

impl LowerHex for Uuid {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.write_str(encode(self.as_bytes(), &mut [0; UUID_STR_LEN], false))
    }
}

impl UpperHex for Uuid {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.write_str(encode(self.as_bytes(), &mut [0; UUID_STR_LEN], true))
    }
}

impl Uuid {
    #[inline]
    pub fn encode_lower<'buf>(&self, buffer: &'buf mut [u8]) -> &'buf mut str {
        encode(self.as_bytes(), buffer, false)
    }

    #[inline]
    pub fn encode_upper<'buf>(&self, buffer: &'buf mut [u8]) -> &'buf mut str {
        encode(self.as_bytes(), buffer, true)
    }
}

pub(crate) fn encode<'a>(data: &[u8], buffer: &'a mut [u8], upper: bool) -> &'a mut str {
    let buf = &mut buffer[..UUID_STR_LEN];
    let dst = buf.as_mut_ptr();

    unsafe {
        ptr::write(dst.cast(), format(data, upper));
        str::from_utf8_unchecked_mut(buf)
    }
}

fn format(data: &[u8], upper: bool) -> [u8; UUID_STR_LEN] {
    let lut = if upper { UPPER_LUT } else { LOWER_LUT };

    let mut dst = [0u8; UUID_STR_LEN];
    let mut i = 0;

    for j in 0..GROUPS.len() {
        let (start, end) = GROUPS[j];

        let mut k = start;
        while k < end {
            let x = data[i];
            i += 1;

            dst[k] = lut[(x >> 4) as usize];
            dst[k + 1] = lut[(x & 0x0f) as usize];
            k += 2;
        }

        if j < GROUPS.len() - 1 {
            dst[end] = b'-';
        }
    }

    dst
}
